public class Galaxy {

/*
**********************************************************************
Class Variables
galtype 0 = point mass
galtype 1 = exponential disk
galtype 2 = singular isothermal spheroid
galtype 3 = isothermal spheroid w/ core
**********************************************************************
*/

   public static final int CORE=3;
   public static final int SINGULAR=2;
   double mass;
   double shape;
   double rcore;
   int galtype;
   boolean blackHole;
   double bheps=0.01;

/*
**********************************************************************
Constructor(s)
**********************************************************************
*/

   Galaxy(double massin, double shapein, double corein, int galtypein,
          boolean bhin) {

      mass=massin;
      shape=shapein;
      rcore=corein;
      galtype=galtypein;
      blackHole=bhin;
   }

   Galaxy() {

      mass=1.0;
      shape=1.0;
      rcore=0.0;
      galtype=1;
      blackHole=false;

   } //Constructor

/*
**********************************************************************
Acceleration
**********************************************************************
*/

   public Triple Acceleration(Triple posin) {

       final double G=1.0;
       double x,y,r2,r,AccMag,InteriorMass;
       Triple acc = new Triple();

       x=posin.x;
       y=posin.y;
       r2=x*x+y*y/(shape*shape);
       r=Math.sqrt(r2);

      switch (galtype) {
      case 0:
        InteriorMass=mass;
        AccMag=-G*InteriorMass/r2;
        acc.x=x*AccMag/r;
        acc.y=y*AccMag/r;
        break;
      case 1:
        InteriorMass=1.-r*Math.exp(-r)-Math.exp(-r);
        AccMag=-G*InteriorMass/r2;
        acc.x=x*AccMag/r;
        acc.y=y*AccMag/r;
        break;
      case 2:
        acc.x=-(x/r2);
        acc.y=-(y/r2/(shape*shape));
        break;
      case 3:
        acc.x=-(x/(rcore*rcore+r2));
        acc.y=-(y/(rcore*rcore+r2)/(shape*shape));
        break;
      }
      if (blackHole) {
	  r=Math.sqrt(x*x+y*y+bheps*bheps);
          acc.x=acc.x-x*0.01*Math.pow(r,-3);
          acc.y=acc.y-y*0.01*Math.pow(r,-3);
      }

      return acc;

   }//Acceleration

/*
**********************************************************************
Potential   
**********************************************************************
*/

   public double Potential(double xin, double yin) {

      final double G=1.0;
      double x,y,r,pot;

      pot=0;
      x=xin;
      y=yin;
      r=Math.sqrt(x*x+y*y);

      switch (galtype) {
      case 0:
        pot=-G*mass/r;
        break;
      case 1:
	  pot=-G*(1.-r*Math.exp(-r)-Math.exp(-r))/r;
        break;
      case 2:
	pot=0.5*Math.log(x*x+y*y/(shape*shape));
        break;
      case 3:
	pot=0.5*Math.log(rcore*rcore+x*x+y*y/(shape*shape));
        break;
      }
      if (blackHole) pot=pot-G*0.01/Math.sqrt(r*r+bheps*bheps);

      return pot;

   }//Potential

/*
**********************************************************************
initVx   
**********************************************************************
*/

   public double initVx(double yin, double vyin, double ein) {

      final double G=1.0;
      double x,y,vy,vx,rad,energy;

      x=0;
      y=yin;
      vy=vyin;
      energy=ein;

      rad=2*(energy-Potential(x,y))-vy*vy;
      if (rad>0)
        vx=Math.sqrt(rad);
      else
        vx=999.0;

      return vx;

   }//initVx

/*
**********************************************************************
maxRadius [solves Pot[rmax]=E]
**********************************************************************
*/

   public double maxRadius(double ein) {

      final double G=1.0;
      double energy,etest,rmax;

      energy=ein;

      rmax=1;
      switch (galtype) {
      case 0:
        rmax=-G*mass/energy;
        break;
      case 1:
        etest=-999.0;
        rmax=0.01;
        while (etest<energy) {
          etest=G*(1-rmax*Math.exp(-rmax)-Math.exp(-rmax))/rmax;
          rmax=rmax+0.01;
        }
        break;
      case 2:
        rmax=Math.exp(+energy);
        break;
      case 3:
        rmax=Math.sqrt(Math.exp(2*energy)-rcore*rcore);
        break;
      }

      return rmax;

   }//maxRadius

/*
**********************************************************************
maxVel [solves v=sqrt(2*(E-Pot))]
**********************************************************************
*/

   public double maxVel(double xin, double yin, double ein) {

      final double G=1.0;
      double vmax;
      double x,y,energy;

      x=xin;
      y=yin;
      energy=ein;

      vmax=1;
      switch (galtype) {
      case 0:
        vmax=Math.sqrt(2*(energy-Potential(x,y)));
        break;
      case 1:
        vmax=Math.sqrt(2*(energy-Potential(x,y)));
        break;
      case 2:
        vmax=Math.sqrt(2*(energy-Potential(x,y)));
        break;
      case 3:
        vmax=Math.sqrt(2*(energy-Potential(x,y)));
        break;
      }

      return vmax;

   }//escapeVel

/*
**********************************************************************
RK4 - integrates using 4th order runge kutta
**********************************************************************
*/
    public void RK4(Triple pos, Triple vel, double dt) {

        Triple z = new Triple();
        Triple p = new Triple();
        Triple q = new Triple();

        Triple zv = new Triple();
        Triple pv = new Triple();
        Triple qv = new Triple();

        z.x=pos.x;
        z.y=pos.y;
        zv.x=vel.x;
        zv.y=vel.y;

        p=Triple.SMultTriple(dt, zv);
        pv=Triple.SMultTriple(dt, Acceleration(z));

        z=Triple.AddTriple(z,Triple.SMultTriple(0.5,p));
        zv=Triple.AddTriple(zv,Triple.SMultTriple(0.5,pv));

        p.copy(q);
        pv.copy(qv);

        p=Triple.SMultTriple(dt, zv);
        pv=Triple.SMultTriple(dt, Acceleration(z));

        z.x=z.x+0.5*p.x-0.5*q.x;
        z.y=z.y+0.5*p.y-0.5*q.y;
        zv.x=zv.x+0.5*pv.x-0.5*qv.x;
        zv.y=zv.y+0.5*pv.y-0.5*qv.y;

        q=Triple.SDivTriple(q,6.0);
        qv=Triple.SDivTriple(qv,6.0);

        p.x=dt*zv.x-0.5*p.x;
        p.y=dt*zv.y-0.5*p.y;
        pv = Triple.AddTriple(Triple.SMultTriple(dt,Acceleration(z)),
                               Triple.SMultTriple(-0.5,pv));

        z=Triple.AddTriple(z,p);
        zv=Triple.AddTriple(zv,pv);

        q=Triple.SubTriple(q,p);
        qv=Triple.SubTriple(qv,pv);

        p = Triple.AddTriple(Triple.SMultTriple(dt,zv),
                               Triple.SMultTriple(2.0,p));
        pv = Triple.AddTriple(Triple.SMultTriple(dt,Acceleration(z)),
                               Triple.SMultTriple(2.0,pv));

        pos.x=z.x+q.x+p.x/6.0;
        pos.y=z.y+q.y+p.y/6.0;
        vel.x=zv.x+qv.x+pv.x/6.0;
        vel.y=zv.y+qv.y+pv.y/6.0;

    }//RK4

}
